/*
   Copyright (C) 2006,2007,2008,2009,2010 Roger While

   This file is part of OpenCOBOL.

   The OpenCOBOL runtime library is free software: you can redistribute it
   and/or modify it under the terms of the GNU Lesser General Public License
   as published by the Free Software Foundation, either version 3 of the
   License, or (at your option) any later version.

   OpenCOBOL is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with OpenCOBOL.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef COB_CODEGEN_H
#define COB_CODEGEN_H

#define	COB_LD_ST2(x,y)		optim_memcpy (x, y, 2);
#define	COB_LD_ST3(x,y)		optim_memcpy (x, y, 3);
#define	COB_LD_ST4(x,y)		optim_memcpy (x, y, 4);
#define	COB_LD_ST5(x,y)		optim_memcpy (x, y, 5);
#define	COB_LD_ST6(x,y)		optim_memcpy (x, y, 6);
#define	COB_LD_ST7(x,y)		optim_memcpy (x, y, 7);
#define	COB_LD_ST8(x,y)		optim_memcpy (x, y, 8);


#ifdef	COB_LOCAL_MALLOC
static void *
cob_local_malloc (const size_t size)
{
	void	*mptr;

	mptr = calloc (1, size);
	if (unlikely(!mptr)) {
		cob_fatal_error (COB_FERROR_MEMORY);
	}
	return mptr;
}
#endif

#ifdef	COB_SET_SCREEN
static void
cob_set_screen (struct cob_screen *s, struct cob_screen *next,
		struct cob_screen *child, cob_field *field, cob_field *value,
		cob_field *line, cob_field *column,
		cob_field *foreg, cob_field *backg, cob_field *prompt,
		const int type, const int occurs, const int attr)
{
	s->next = next;
	s->child = child;
	s->field = field;
	s->value = value;
	s->line = line;
	s->column = column;
	s->foreg = foreg;
	s->backg = backg;
	s->prompt = prompt;
	s->type = type;
	s->occurs = occurs;
	s->attr = attr;
}
#endif

#ifdef	COB_DECIMAL_SET_INT
static COB_INLINE void
cob_decimal_set_int (cob_decimal *d, const int n)
{
	mpz_set_si (d->value, n);
	d->scale = 0;
}
#endif

#ifdef	COB_DECIMAL_SET_UINT
static COB_INLINE void
cob_decimal_set_uint (cob_decimal *d, const unsigned int n)
{
	mpz_set_ui (d->value, n);
	d->scale = 0;
}
#endif

#ifdef	COB_POINTER_MANIP
static void
cob_pointer_manip (cob_field *f1, cob_field *f2, const unsigned int addsub)
{
	unsigned char   *tmptr;
	memcpy (&tmptr, f1->data, sizeof(void *));
	if (addsub) {
		tmptr -= cob_get_int (f2);
	} else {
		tmptr += cob_get_int (f2);
	}
	memcpy (f1->data, &tmptr, sizeof(void *));
}
#endif

#ifdef	COB_GET_NUMDISP
static int
cob_get_numdisp (const unsigned char *data, const size_t size)
{
	size_t	n;
	int     retval;

	retval = 0;
	for (n = 0; n < size; ++n, ++data) {
		if (likely(*data >= (unsigned char)'0' &&
		    *data <= (unsigned char)'9')) {
			retval *= 10;
			retval += (*data - (unsigned char)'0');
		}
	}
	return retval;
}
#endif

#ifdef	COB_CMP_PACKED_INT
static int
cob_cmp_packed_int (const cob_field *f, const int n)
{
	unsigned char	*p;
	size_t		size;
	int		val;

	val = 0;
	p = f->data;
	for (size = 0; size < f->size - 1; ++size, ++p) {
		val *= 10;
		val += *p >> 4;
		val *= 10;
		val += *p & 0x0f;
	}
	val *= 10;
	val += *p >> 4;
	if ((*p & 0x0f) == 0x0d) {
		val = -val;
	}
	return (val < n) ? -1 : (val > n);
}
#endif

#ifdef	COB_GET_PACKED_INT
static int
cob_get_packed_int (const cob_field *f)
{
	unsigned char	*p;
	size_t		size;
	int		val = 0;

	p = f->data;
	for (size = 0; size < f->size - 1; ++size, ++p) {
		val *= 10;
		val += *p >> 4;
		val *= 10;
		val += *p & 0x0f;
	}
	val *= 10;
	val += *p >> 4;
	if ((*p & 0x0f) == 0x0d) {
		val = -val;
	}
	return val;
}
#endif

#ifdef	COB_ADD_PACKED_INT
static int
cob_add_packed_int (cob_field *f, const int val)
{
	unsigned char	*p;
	size_t		size;
	int		carry = 0;
	int		n;
	int		inc;

	if (val == 0) {
		return 0;
	}
	p = f->data + f->size - 1;
	if ((*p & 0x0f) == 0x0d) {
		if (val > 0) {
			return cob_add_int (f, val, 0);
		}
		n = -val;
	} else {
		if (val < 0) {
			return cob_add_int (f, val, 0);
		}
		n = val;
	}
	inc = (*p >> 4) + (n % 10);
	n /= 10;
	carry = inc / 10;
	*p = ((inc % 10) << 4) | (*p & 0x0f);
	p--;

	for (size = 0; size < f->size - 1; ++size, --p) {
		if (!carry && !n) {
			break;
		}
		inc = ((*p >> 4) * 10) + (*p & 0x0f) + carry + (n % 100);
		carry = inc / 100;
		n /= 100;
		inc %= 100;
		*p = ((inc / 10) << 4) | (inc % 10);
	}
	return 0;
}
#endif

/* Aligned variants */

/* Aligned compares */

#ifdef	COB_CMP_ALIGN_U16
static COB_INLINE COB_A_INLINE int
cob_cmp_align_u16 (const unsigned char *p, const int n)
{
	unsigned short	val;

	if (unlikely(n < 0)) {
		return 1;
	}
	val = *(unsigned short __unaligned *)p;
	return (val < n) ? -1 : (val > n);
}
#endif

#ifdef	COB_CMP_ALIGN_S16
static COB_INLINE COB_A_INLINE int
cob_cmp_align_s16 (const unsigned char *p, const int n)
{
	short	val;

	val = *(short __unaligned *)p;
	return (val < n) ? -1 : (val > n);
}
#endif

#ifdef	COB_CMP_ALIGN_U32
static COB_INLINE COB_A_INLINE int
cob_cmp_align_u32 (const unsigned char *p, const int n)
{
	unsigned int	val;

	if (unlikely(n < 0)) {
		return 1;
	}
	val = *(unsigned int __unaligned *)p;
	return (val < n) ? -1 : (val > n);
}
#endif

#ifdef	COB_CMP_ALIGN_S32
static COB_INLINE COB_A_INLINE int
cob_cmp_align_s32 (const unsigned char *p, const int n)
{
	int	val;

	val = *(int __unaligned *)p;
	return (val < n) ? -1 : (val > n);
}
#endif

#ifdef	COB_CMP_ALIGN_U64
static COB_INLINE COB_A_INLINE int
cob_cmp_align_u64 (const unsigned char *p, const int n)
{
	cob_u64_t	val;

	if (unlikely(n < 0)) {
		return 1;
	}
	val = *(cob_u64_t __unaligned *)p;
	return (val < n) ? -1 : (val > n);
}
#endif

#ifdef	COB_CMP_ALIGN_S64
static COB_INLINE COB_A_INLINE int
cob_cmp_align_s64 (const unsigned char *p, const int n)
{
	cob_s64_t	val;

	val = *(cob_s64_t __unaligned *)p;
	return (val < n) ? -1 : (val > n);
}
#endif

/* Aligned adds */

#ifdef	COB_ADD_ALIGN_U16
static COB_INLINE COB_A_INLINE void
cob_add_align_u16 (unsigned char *p, const int val)
{
	*(unsigned short __unaligned *)p += val;
}
#endif

#ifdef	COB_ADD_ALIGN_S16
static COB_INLINE COB_A_INLINE void
cob_add_align_s16 (unsigned char *p, const int val)
{
	*(short __unaligned *)p += val;
}
#endif

#ifdef	COB_ADD_ALIGN_U32
static COB_INLINE COB_A_INLINE void
cob_add_align_u32 (unsigned char *p, const int val)
{
	*(unsigned int __unaligned *)p += val;
}
#endif

#ifdef	COB_ADD_ALIGN_S32
static COB_INLINE COB_A_INLINE void
cob_add_align_s32 (unsigned char *p, const int val)
{
	*(int __unaligned *)p += val;
}
#endif

#ifdef	COB_ADD_ALIGN_U64
static COB_INLINE COB_A_INLINE void
cob_add_align_u64 (unsigned char *p, const int val)
{
	*(cob_u64_t __unaligned *)p += val;
}
#endif

#ifdef	COB_ADD_ALIGN_S64
static COB_INLINE COB_A_INLINE void
cob_add_align_s64 (unsigned char *p, const int val)
{
	*(cob_s64_t __unaligned *)p += val;
}
#endif

/* Aligned subtracts */

#ifdef	COB_SUB_ALIGN_U16
static COB_INLINE COB_A_INLINE void
cob_sub_align_u16 (unsigned char *p, const int val)
{
	*(unsigned short __unaligned *)p -= val;
}
#endif

#ifdef	COB_SUB_ALIGN_S16
static COB_INLINE COB_A_INLINE void
cob_sub_align_s16 (unsigned char *p, const int val)
{
	*(short __unaligned *)p -= val;
}
#endif

#ifdef	COB_SUB_ALIGN_U32
static COB_INLINE COB_A_INLINE void
cob_sub_align_u32 (unsigned char *p, const int val)
{
	*(unsigned int __unaligned *)p -= val;
}
#endif

#ifdef	COB_SUB_ALIGN_S32
static COB_INLINE COB_A_INLINE void
cob_sub_align_s32 (unsigned char *p, const int val)
{
	*(int __unaligned *)p -= val;
}
#endif

#ifdef	COB_SUB_ALIGN_U64
static COB_INLINE COB_A_INLINE void
cob_sub_align_u64 (unsigned char *p, const int val)
{
	*(cob_u64_t __unaligned *)p -= val;
}
#endif

#ifdef	COB_SUB_ALIGN_S64
static COB_INLINE COB_A_INLINE void
cob_sub_align_s64 (unsigned char *p, const int val)
{
	*(cob_s64_t __unaligned *)p -= val;
}
#endif

#ifdef	COB_CMPSWP_ALIGN_U16
static COB_INLINE COB_A_INLINE int
cob_cmpswp_align_u16 (const unsigned char *p, const int n)
{
	unsigned short	val;

	if (unlikely(n < 0)) {
		return 1;
	}
	val = COB_BSWAP_16 (*(unsigned short __unaligned *)p);
	return (val < n) ? -1 : (val > n);
}
#endif

#ifdef	COB_CMPSWP_ALIGN_S16
static COB_INLINE COB_A_INLINE int
cob_cmpswp_align_s16 (const unsigned char *p, const int n)
{
	short	val;

	val = COB_BSWAP_16 (*(short __unaligned *)p);
	return (val < n) ? -1 : (val > n);
}
#endif

#ifdef	COB_CMPSWP_ALIGN_U32
static COB_INLINE COB_A_INLINE int
cob_cmpswp_align_u32 (const unsigned char *p, const int n)
{
	unsigned int	val;

	if (unlikely(n < 0)) {
		return 1;
	}
	val = COB_BSWAP_32 (*(unsigned int __unaligned *)p);
	return (val < n) ? -1 : (val > n);
}
#endif

#ifdef	COB_CMPSWP_ALIGN_S32
static COB_INLINE COB_A_INLINE int
cob_cmpswp_align_s32 (const unsigned char *p, const int n)
{
	int	val;

	val = COB_BSWAP_32 (*(int __unaligned *)p);
	return (val < n) ? -1 : (val > n);
}
#endif

#ifdef	COB_CMPSWP_ALIGN_U64
static COB_INLINE COB_A_INLINE int
cob_cmpswp_align_u64 (const unsigned char *p, const int n)
{
	cob_u64_t	val;

	if (unlikely(n < 0)) {
		return 1;
	}
	val = COB_BSWAP_64 (*(cob_u64_t __unaligned *)p);
	return (val < n) ? -1 : (val > n);
}
#endif

#ifdef	COB_CMPSWP_ALIGN_S64
static COB_INLINE COB_A_INLINE int
cob_cmpswp_align_s64 (const unsigned char *p, const int n)
{
	cob_s64_t	val;

	val = COB_BSWAP_64 (*(cob_s64_t __unaligned *)p);
	return (val < n) ? -1 : (val > n);
}
#endif

/* Binary compare */

#ifdef	COB_CMP_U8
static COB_INLINE COB_A_INLINE int
cob_cmp_u8 (const unsigned char *p, const int n)
{
	if (unlikely(n < 0)) {
		return 1;
	}
	return (*p < n) ? -1 : (*p > n);
}
#endif

#ifdef	COB_CMP_S8
static COB_INLINE COB_A_INLINE int
cob_cmp_s8 (const unsigned char *p, const int n)
{
	return (*(const signed char *)p < n) ? -1 : (*(const signed char *)p > n);
}
#endif

#ifdef	COB_CMP_U16
static COB_INLINE COB_A_INLINE int
cob_cmp_u16 (const unsigned char *p, const int n)
{
#ifndef COB_ALLOW_UNALIGNED
	unsigned char	*x;
#endif
	unsigned short	val;

	if (unlikely(n < 0)) {
		return 1;
	}
#ifdef	COB_ALLOW_UNALIGNED
	val = *(const unsigned short __unaligned *)p;
#else
	x = (unsigned char *)&val;
	COB_LD_ST2 (x, p);
#endif
	return (val < n) ? -1 : (val > n);
}
#endif

#ifdef	COB_CMP_S16
static COB_INLINE COB_A_INLINE int
cob_cmp_s16 (const unsigned char *p, const int n)
{
	short	val;

#ifdef	COB_ALLOW_UNALIGNED
	val = *(const short __unaligned *)p;
#else
	unsigned char	*x;

	x = (unsigned char *)&val;
	COB_LD_ST2 (x, p);
#endif
	return (val < n) ? -1 : (val > n);
}
#endif

#ifdef	COB_CMP_U24
static COB_INLINE COB_A_INLINE int
cob_cmp_u24 (const unsigned char *p, const int n)
{
	unsigned char	*x;
	unsigned int	val = 0;

	if (unlikely(n < 0)) {
		return 1;
	}
#ifdef	WORDS_BIGENDIAN
	x = ((unsigned char *)&val) + 1;
#else
	x = (unsigned char *)&val;
#endif
	COB_LD_ST3 (x, p);
	return (val < n) ? -1 : (val > n);
}
#endif

#ifdef	COB_CMP_S24
static COB_INLINE COB_A_INLINE int
cob_cmp_s24 (const unsigned char *p, const int n)
{
	unsigned char	*x;
	int		val = 0;

#ifdef	WORDS_BIGENDIAN
	x = (unsigned char *)&val;
#else
	x = ((unsigned char *)&val) + 1;
#endif
	COB_LD_ST3 (x, p);
	val >>= 8;	/* Shift with sign */
	return (val < n) ? -1 : (val > n);
}
#endif

#ifdef	COB_CMP_U32
static COB_INLINE COB_A_INLINE int
cob_cmp_u32 (const unsigned char *p, const int n)
{
#ifndef COB_ALLOW_UNALIGNED
	unsigned char	*x;
#endif
	unsigned int	val;

	if (unlikely(n < 0)) {
		return 1;
	}
#ifdef	COB_ALLOW_UNALIGNED
	val = *(const unsigned int __unaligned *)p;
#else
	x = (unsigned char *)&val;
	COB_LD_ST4 (x, p);
#endif
	return (val < n) ? -1 : (val > n);
}
#endif

#ifdef	COB_CMP_S32
static COB_INLINE COB_A_INLINE int
cob_cmp_s32 (const unsigned char *p, const int n)
{
	int	val;

#ifdef	COB_ALLOW_UNALIGNED
	val = *(const int __unaligned *)p;
#else
	unsigned char	*x;

	x = (unsigned char *)&val;
	COB_LD_ST4 (x, p);
#endif
	return (val < n) ? -1 : (val > n);
}
#endif

#ifdef	COB_CMP_U40
static COB_INLINE COB_A_INLINE int
cob_cmp_u40 (const unsigned char *p, const int n)
{
	cob_u64_t		val = 0;
	unsigned char		*x;

	if (unlikely(n < 0)) {
		return 1;
	}
#ifdef	WORDS_BIGENDIAN
	x = ((unsigned char *)&val) + 3;
#else
	x = (unsigned char *)&val;
#endif
	COB_LD_ST5 (x, p);
	return (val < n) ? -1 : (val > n);
}
#endif

#ifdef	COB_CMP_S40
static COB_INLINE COB_A_INLINE int
cob_cmp_s40 (const unsigned char *p, const int n)
{
	cob_s64_t		val = 0;
	unsigned char		*x;

#ifdef	WORDS_BIGENDIAN
	x = (unsigned char *)&val;
#else
	x = ((unsigned char *)&val) + 3;
#endif
	COB_LD_ST5 (x, p);
	val >>= 24;	/* Shift with sign */
	return (val < n) ? -1 : (val > n);
}
#endif

#ifdef	COB_CMP_U48
static COB_INLINE COB_A_INLINE int
cob_cmp_u48 (const unsigned char *p, const int n)
{
	cob_u64_t		val = 0;
	unsigned char		*x;

	if (unlikely(n < 0)) {
		return 1;
	}
#ifdef	WORDS_BIGENDIAN
	x = ((unsigned char *)&val) + 2;
#else
	x = (unsigned char *)&val;
#endif
	COB_LD_ST6 (x, p);
	return (val < n) ? -1 : (val > n);
}
#endif

#ifdef	COB_CMP_S48
static COB_INLINE COB_A_INLINE int
cob_cmp_s48 (const unsigned char *p, const int n)
{
	cob_s64_t		val = 0;
	unsigned char		*x;

#ifdef	WORDS_BIGENDIAN
	x = (unsigned char *)&val;
#else
	x = ((unsigned char *)&val) + 2;
#endif
	COB_LD_ST6 (x, p);
	val >>= 16;	/* Shift with sign */
	return (val < n) ? -1 : (val > n);
}
#endif

#ifdef	COB_CMP_U56
static COB_INLINE COB_A_INLINE int
cob_cmp_u56 (const unsigned char *p, const int n)
{
	cob_u64_t		val = 0;
	unsigned char		*x;

	if (unlikely(n < 0)) {
		return 1;
	}
#ifdef	WORDS_BIGENDIAN
	x = ((unsigned char *)&val) + 1;
#else
	x = (unsigned char *)&val;
#endif
	COB_LD_ST7 (x, p);
	return (val < n) ? -1 : (val > n);
}
#endif

#ifdef	COB_CMP_S56
static COB_INLINE COB_A_INLINE int
cob_cmp_s56 (const unsigned char *p, const int n)
{
	cob_s64_t		val = 0;
	unsigned char		*x;

#ifdef	WORDS_BIGENDIAN
	x = (unsigned char *)&val;
#else
	x = ((unsigned char *)&val) + 1;
#endif
	COB_LD_ST7 (x, p);
	val >>= 8;	/* Shift with sign */
	return (val < n) ? -1 : (val > n);
}
#endif

#ifdef	COB_CMP_U64
static COB_INLINE COB_A_INLINE int
cob_cmp_u64 (const unsigned char *p, const int n)
{
#ifndef COB_ALLOW_UNALIGNED
	unsigned char		*x;
#endif
	cob_u64_t		val;

	if (unlikely(n < 0)) {
		return 1;
	}
#ifdef	COB_ALLOW_UNALIGNED
	val = *(const cob_u64_t __unaligned *)p;
#else
	x = (unsigned char *)&val;
	COB_LD_ST8 (x, p);
#endif
	return (val < n) ? -1 : (val > n);
}
#endif

#ifdef	COB_CMP_S64
static COB_INLINE COB_A_INLINE int
cob_cmp_s64 (const unsigned char *p, const int n)
{
	cob_s64_t		val;

#ifdef	COB_ALLOW_UNALIGNED
	val = *(const cob_s64_t __unaligned *)p;
#else
	unsigned char		*x;

	x = (unsigned char *)&val;
	COB_LD_ST8 (x, p);
#endif
	return (val < n) ? -1 : (val > n);
}
#endif

/* Add/Subtract */

#ifdef	COB_ADD_U8
static COB_INLINE COB_A_INLINE void
cob_add_u8 (unsigned char *p, const int val)
{
	*p += val;
}
#endif

#ifdef	COB_ADD_S8
static COB_INLINE COB_A_INLINE void
cob_add_s8 (unsigned char *p, const int val)
{
	*(signed char *)p += val;
}
#endif

#ifdef	COB_ADD_U16
static COB_INLINE COB_A_INLINE void
cob_add_u16 (unsigned char *p, const int val)
{
#ifdef	COB_ALLOW_UNALIGNED
	*(unsigned short __unaligned *)p += val;
#else
	unsigned char	*x;
	unsigned short	n;

	x = (unsigned char *)&n;
	COB_LD_ST2 (x, p);
	n += val;
	COB_LD_ST2 (p, x);
#endif
}
#endif

#ifdef	COB_ADD_S16
static COB_INLINE COB_A_INLINE void
cob_add_s16 (unsigned char *p, const int val)
{
#ifdef	COB_ALLOW_UNALIGNED
	*(short __unaligned *)p += val;
#else
	unsigned char	*x;
	short		n;

	x = (unsigned char *)&n;
	COB_LD_ST2 (x, p);
	n += val;
	COB_LD_ST2 (p, x);
#endif
}
#endif

#ifdef	COB_ADD_U24
static COB_INLINE COB_A_INLINE void
cob_add_u24 (unsigned char *p, const int val)
{
	unsigned char	*x;
	unsigned int	n = 0;

#ifdef	WORDS_BIGENDIAN
	x = ((unsigned char *)&n) + 1;
#else
	x = (unsigned char *)&n;
#endif
	COB_LD_ST3 (x, p);
	n += val;
	COB_LD_ST3 (p, x);
}
#endif

#ifdef	COB_ADD_S24
static COB_INLINE COB_A_INLINE void
cob_add_s24 (unsigned char *p, const int val)
{
	unsigned char	*x;
	int		n = 0;

#ifdef	WORDS_BIGENDIAN
	x = (unsigned char *)&n;
#else
	x = ((unsigned char *)&n) + 1;
#endif
	COB_LD_ST3 (x, p);
	n >>= 8;	/* Shift with sign */
	n += val;
#ifdef	WORDS_BIGENDIAN
	x = ((unsigned char *)&n) + 1;
#else
	x = (unsigned char *)&n;
#endif
	COB_LD_ST3 (p, x);
}
#endif

#ifdef	COB_ADD_U32
static COB_INLINE COB_A_INLINE void
cob_add_u32 (unsigned char *p, const int val)
{
#ifdef	COB_ALLOW_UNALIGNED
	*(unsigned int __unaligned *)p += val;
#else
	unsigned char	*x;
	unsigned int	n;

	x = (unsigned char *)&n;
	COB_LD_ST4 (x, p);
	n += val;
	COB_LD_ST4 (p, x);
#endif
}
#endif

#ifdef	COB_ADD_S32
static COB_INLINE COB_A_INLINE void
cob_add_s32 (unsigned char *p, const int val)
{
#ifdef	COB_ALLOW_UNALIGNED
	*(int __unaligned *)p += val;
#else
	unsigned char	*x;
	int		n;

	x = (unsigned char *)&n;
	COB_LD_ST4 (x, p);
	n += val;
	COB_LD_ST4 (p, x);
#endif
}
#endif

#ifdef	COB_ADD_U40
static COB_INLINE COB_A_INLINE void
cob_add_u40 (unsigned char *p, const int val)
{
	cob_u64_t		n = 0;
	unsigned char		*x;

#ifdef	WORDS_BIGENDIAN
	x = ((unsigned char *)&n) + 3;
#else
	x = (unsigned char *)&n;
#endif
	COB_LD_ST5 (x, p);
	n += val;
	COB_LD_ST5 (p, x);
}
#endif

#ifdef	COB_ADD_S40
static COB_INLINE COB_A_INLINE void
cob_add_s40 (unsigned char *p, const int val)
{
	cob_s64_t		n = 0;
	unsigned char		*x;

#ifdef	WORDS_BIGENDIAN
	x = (unsigned char *)&n;
#else
	x = ((unsigned char *)&n) + 3;
#endif
	COB_LD_ST5 (x, p);
	n >>= 24;	/* Shift with sign */
	n += val;
#ifdef	WORDS_BIGENDIAN
	x = ((unsigned char *)&n) + 3;
#else
	x = (unsigned char *)&n;
#endif
	COB_LD_ST5 (p, x);
}
#endif

#ifdef	COB_ADD_U48
static COB_INLINE COB_A_INLINE void
cob_add_u48 (unsigned char *p, const int val)
{
	cob_u64_t		n = 0;
	unsigned char		*x;

#ifdef	WORDS_BIGENDIAN
	x = ((unsigned char *)&n) + 2;
#else
	x = (unsigned char *)&n;
#endif
	COB_LD_ST6 (x, p);
	n += val;
	COB_LD_ST6 (p, x);
}
#endif

#ifdef	COB_ADD_S48
static COB_INLINE COB_A_INLINE void
cob_add_s48 (unsigned char *p, const int val)
{
	cob_s64_t		n = 0;
	unsigned char		*x;

#ifdef	WORDS_BIGENDIAN
	x = (unsigned char *)&n;
#else
	x = ((unsigned char *)&n) + 2;
#endif
	COB_LD_ST6 (x, p);
	n >>= 16;	/* Shift with sign */
	n += val;
#ifdef	WORDS_BIGENDIAN
	x = ((unsigned char *)&n) + 2;
#else
	x = (unsigned char *)&n;
#endif
	COB_LD_ST6 (p, x);
}
#endif

#ifdef	COB_ADD_U56
static COB_INLINE COB_A_INLINE void
cob_add_u56 (unsigned char *p, const int val)
{
	cob_u64_t		n = 0;
	unsigned char		*x;

#ifdef	WORDS_BIGENDIAN
	x = ((unsigned char *)&n) + 1;
#else
	x = (unsigned char *)&n;
#endif
	COB_LD_ST7 (x, p);
	n += val;
	COB_LD_ST7 (p, x);
}
#endif

#ifdef	COB_ADD_S56
static COB_INLINE COB_A_INLINE void
cob_add_s56 (unsigned char *p, const int val)
{
	cob_s64_t		n = 0;
	unsigned char		*x;

#ifdef	WORDS_BIGENDIAN
	x = (unsigned char *)&n;
#else
	x = ((unsigned char *)&n) + 1;
#endif
	COB_LD_ST7 (x, p);
	n >>= 8;	/* Shift with sign */
	n += val;
#ifdef	WORDS_BIGENDIAN
	x = ((unsigned char *)&n) + 1;
#else
	x = (unsigned char *)&n;
#endif
	COB_LD_ST7 (p, x);
}
#endif

#ifdef	COB_ADD_U64
static COB_INLINE COB_A_INLINE void
cob_add_u64 (unsigned char *p, const int val)
{
#ifdef	COB_ALLOW_UNALIGNED
	*(cob_u64_t __unaligned *)p += val;
#else
	unsigned char		*x;
	cob_u64_t		n;

	x = (unsigned char *)&n;
	COB_LD_ST8 (x, p);
	n += val;
	COB_LD_ST8 (p, x);
#endif
}
#endif

#ifdef	COB_ADD_S64
static COB_INLINE COB_A_INLINE void
cob_add_s64 (unsigned char *p, const int val)
{
#ifdef	COB_ALLOW_UNALIGNED
	*(cob_s64_t __unaligned *)p += val;
#else
	unsigned char		*x;
	cob_s64_t		n;

	x = (unsigned char *)&n;
	COB_LD_ST8 (x, p);
	n += val;
	COB_LD_ST8 (p, x);
#endif
}
#endif

#ifdef	COB_SUB_U8
static COB_INLINE COB_A_INLINE void
cob_sub_u8 (unsigned char *p, const int val)
{
	*p -= val;
}
#endif

#ifdef	COB_SUB_S8
static COB_INLINE COB_A_INLINE void
cob_sub_s8 (unsigned char *p, const int val)
{
	*(signed char *)p -= val;
}
#endif

#ifdef	COB_SUB_U16
static COB_INLINE COB_A_INLINE void
cob_sub_u16 (unsigned char *p, const int val)
{
#ifdef	COB_ALLOW_UNALIGNED
	*(unsigned short __unaligned *)p -= val;
#else
	unsigned char	*x;
	unsigned short	n;

	x = (unsigned char *)&n;
	COB_LD_ST2 (x, p);
	n -= val;
	COB_LD_ST2 (p, x);
#endif
}
#endif

#ifdef	COB_SUB_S16
static COB_INLINE COB_A_INLINE void
cob_sub_s16 (unsigned char *p, const int val)
{
#ifdef	COB_ALLOW_UNALIGNED
	*(short __unaligned *)p -= val;
#else
	unsigned char	*x;
	short		n;

	x = (unsigned char *)&n;
	COB_LD_ST2 (x, p);
	n -= val;
	COB_LD_ST2 (p, x);
#endif
}
#endif

#ifdef	COB_SUB_U24
static COB_INLINE COB_A_INLINE void
cob_sub_u24 (unsigned char *p, const int val)
{
	unsigned char	*x;
	unsigned int	n = 0;

#ifdef	WORDS_BIGENDIAN
	x = ((unsigned char *)&n) + 1;
#else
	x = (unsigned char *)&n;
#endif
	COB_LD_ST3 (x, p);
	n -= val;
	COB_LD_ST3 (p, x);
}
#endif

#ifdef	COB_SUB_S24
static COB_INLINE COB_A_INLINE void
cob_sub_s24 (unsigned char *p, const int val)
{
	unsigned char	*x;
	int		n = 0;

#ifdef	WORDS_BIGENDIAN
	x = (unsigned char *)&n;
#else
	x = ((unsigned char *)&n) + 1;
#endif
	COB_LD_ST3 (x, p);
	n >>= 8;	/* Shift with sign */
	n -= val;
#ifdef	WORDS_BIGENDIAN
	x = ((unsigned char *)&n) + 1;
#else
	x = (unsigned char *)&n;
#endif
	COB_LD_ST3 (p, x);
}
#endif

#ifdef	COB_SUB_U32
static COB_INLINE COB_A_INLINE void
cob_sub_u32 (unsigned char *p, const int val)
{
#ifdef	COB_ALLOW_UNALIGNED
	*(unsigned int __unaligned *)p -= val;
#else
	unsigned char	*x;
	unsigned int	n;

	x = (unsigned char *)&n;
	COB_LD_ST4 (x, p);
	n -= val;
	COB_LD_ST4 (p, x);
#endif
}
#endif

#ifdef	COB_SUB_S32
static COB_INLINE COB_A_INLINE void
cob_sub_s32 (unsigned char *p, const int val)
{
#ifdef	COB_ALLOW_UNALIGNED
	*(int __unaligned *)p -= val;
#else
	unsigned char	*x;
	int		n;

	x = (unsigned char *)&n;
	COB_LD_ST4 (x, p);
	n -= val;
	COB_LD_ST4 (p, x);
#endif
}
#endif

#ifdef	COB_SUB_U40
static COB_INLINE COB_A_INLINE void
cob_sub_u40 (unsigned char *p, const int val)
{
	cob_u64_t		n = 0;
	unsigned char		*x;

#ifdef	WORDS_BIGENDIAN
	x = ((unsigned char *)&n) + 3;
#else
	x = (unsigned char *)&n;
#endif
	COB_LD_ST5 (x, p);
	n -= val;
	COB_LD_ST5 (p, x);
}
#endif

#ifdef	COB_SUB_S40
static COB_INLINE COB_A_INLINE void
cob_sub_s40 (unsigned char *p, const int val)
{
	cob_s64_t		n = 0;
	unsigned char		*x;

#ifdef	WORDS_BIGENDIAN
	x = (unsigned char *)&n;
#else
	x = ((unsigned char *)&n) + 3;
#endif
	COB_LD_ST5 (x, p);
	n >>= 24;	/* Shift with sign */
	n -= val;
#ifdef	WORDS_BIGENDIAN
	x = ((unsigned char *)&n) + 3;
#else
	x = (unsigned char *)&n;
#endif
	COB_LD_ST5 (p, x);
}
#endif

#ifdef	COB_SUB_U48
static COB_INLINE COB_A_INLINE void
cob_sub_u48 (unsigned char *p, const int val)
{
	cob_u64_t		n = 0;
	unsigned char		*x;

#ifdef	WORDS_BIGENDIAN
	x = ((unsigned char *)&n) + 2;
#else
	x = (unsigned char *)&n;
#endif
	COB_LD_ST6 (x, p);
	n -= val;
	COB_LD_ST6 (p, x);
}
#endif

#ifdef	COB_SUB_S48
static COB_INLINE COB_A_INLINE void
cob_sub_s48 (unsigned char *p, const int val)
{
	cob_s64_t		n = 0;
	unsigned char		*x;

#ifdef	WORDS_BIGENDIAN
	x = (unsigned char *)&n;
#else
	x = ((unsigned char *)&n) + 2;
#endif
	COB_LD_ST6 (x, p);
	n >>= 16;	/* Shift with sign */
	n -= val;
#ifdef	WORDS_BIGENDIAN
	x = ((unsigned char *)&n) + 2;
#else
	x = (unsigned char *)&n;
#endif
	COB_LD_ST6 (p, x);
}
#endif

#ifdef	COB_SUB_U56
static COB_INLINE COB_A_INLINE void
cob_sub_u56 (unsigned char *p, const int val)
{
	cob_u64_t		n = 0;
	unsigned char		*x;

#ifdef	WORDS_BIGENDIAN
	x = ((unsigned char *)&n) + 1;
#else
	x = (unsigned char *)&n;
#endif
	COB_LD_ST7 (x, p);
	n -= val;
	COB_LD_ST7 (p, x);
}
#endif

#ifdef	COB_SUB_S56
static COB_INLINE COB_A_INLINE void
cob_sub_s56 (unsigned char *p, const int val)
{
	cob_s64_t		n = 0;
	unsigned char		*x;

#ifdef	WORDS_BIGENDIAN
	x = (unsigned char *)&n;
#else
	x = ((unsigned char *)&n) + 1;
#endif
	COB_LD_ST7 (x, p);
	n >>= 8;	/* Shift with sign */
	n -= val;
#ifdef	WORDS_BIGENDIAN
	x = ((unsigned char *)&n) + 1;
#else
	x = (unsigned char *)&n;
#endif
	COB_LD_ST7 (p, x);
}
#endif

#ifdef	COB_SUB_U64
static COB_INLINE COB_A_INLINE void
cob_sub_u64 (unsigned char *p, const int val)
{
#ifdef	COB_ALLOW_UNALIGNED
	*(cob_u64_t __unaligned *)p -= val;
#else
	unsigned char		*x;
	cob_u64_t	n;

	x = (unsigned char *)&n;
	COB_LD_ST8 (x, p);
	n -= val;
	COB_LD_ST8 (p, x);
#endif
}
#endif

#ifdef	COB_SUB_S64
static COB_INLINE COB_A_INLINE void
cob_sub_s64 (unsigned char *p, const int val)
{
#ifdef	COB_ALLOW_UNALIGNED
	*(cob_s64_t __unaligned *)p -= val;
#else
	unsigned char		*x;
	cob_s64_t		n;

	x = (unsigned char *)&n;
	COB_LD_ST8 (x, p);
	n -= val;
	COB_LD_ST8 (p, x);
#endif
}
#endif

/* Binary swapped compare */
#ifdef	COB_CMPSWP_U16
static COB_INLINE COB_A_INLINE int
cob_cmpswp_u16 (const unsigned char *p, const int n)
{
#ifndef COB_ALLOW_UNALIGNED
	unsigned char	*x;
#endif
	unsigned short	val;

	if (unlikely(n < 0)) {
		return 1;
	}
#ifdef	COB_ALLOW_UNALIGNED
	val = COB_BSWAP_16 (*(unsigned short __unaligned *)p);
#else
	x = (unsigned char *)&val;
	COB_LD_ST2 (x, p);
	val = COB_BSWAP_16 (val);
#endif
	return (val < n) ? -1 : (val > n);
}
#endif

#ifdef	COB_CMPSWP_S16
static COB_INLINE COB_A_INLINE int
cob_cmpswp_s16 (const unsigned char *p, const int n)
{
	short	val;

#ifdef	COB_ALLOW_UNALIGNED
	val = COB_BSWAP_16 (*(short __unaligned *)p);
#else
	unsigned char	*x;

	x = (unsigned char *)&val;
	COB_LD_ST2 (x, p);
	val = COB_BSWAP_16 (val);
#endif
	return (val < n) ? -1 : (val > n);
}
#endif

#ifdef	COB_CMPSWP_U24
static COB_INLINE COB_A_INLINE int
cob_cmpswp_u24 (const unsigned char *p, const int n)
{
	unsigned char	*x;
	unsigned int	val = 0;

	if (unlikely(n < 0)) {
		return 1;
	}
	x = ((unsigned char *)&val) + 1;
	COB_LD_ST3 (x, p);
	val = COB_BSWAP_32 (val);
	return (val < n) ? -1 : (val > n);
}
#endif

#ifdef	COB_CMPSWP_S24
static COB_INLINE COB_A_INLINE int
cob_cmpswp_s24 (const unsigned char *p, const int n)
{
	unsigned char	*x;
	int		val = 0;

	x = (unsigned char *)&val;
	COB_LD_ST3 (x, p);
	val = COB_BSWAP_32 (val);
	val >>= 8;	/* Shift with sign */
	return (val < n) ? -1 : (val > n);
}
#endif

#ifdef	COB_CMPSWP_U32
static COB_INLINE COB_A_INLINE int
cob_cmpswp_u32 (const unsigned char *p, const int n)
{
#ifndef COB_ALLOW_UNALIGNED
	unsigned char	*x;
#endif
	unsigned int	val;

	if (unlikely(n < 0)) {
		return 1;
	}
#ifdef	COB_ALLOW_UNALIGNED
	val = COB_BSWAP_32 (*(const unsigned int __unaligned *)p);
#else
	x = (unsigned char *)&val;
	COB_LD_ST4 (x, p);
	val = COB_BSWAP_32 (val);
#endif
	return (val < n) ? -1 : (val > n);
}
#endif

#ifdef	COB_CMPSWP_S32
static COB_INLINE COB_A_INLINE int
cob_cmpswp_s32 (const unsigned char *p, const int n)
{
	int	val;

#ifdef	COB_ALLOW_UNALIGNED
	val = COB_BSWAP_32 (*(const int __unaligned *)p);
#else
	unsigned char	*x;

	x = (unsigned char *)&val;
	COB_LD_ST4 (x, p);
	val = COB_BSWAP_32 (val);
#endif
	return (val < n) ? -1 : (val > n);
}
#endif

#ifdef	COB_CMPSWP_U40
static COB_INLINE COB_A_INLINE int
cob_cmpswp_u40 (const unsigned char *p, const int n)
{
	cob_u64_t	val = 0;
	unsigned char	*x;

	if (unlikely(n < 0)) {
		return 1;
	}
	x = ((unsigned char *)&val) + 3;
	COB_LD_ST5 (x, p);
	val = COB_BSWAP_64 (val);
	return (val < n) ? -1 : (val > n);
}
#endif

#ifdef	COB_CMPSWP_S40
static COB_INLINE COB_A_INLINE int
cob_cmpswp_s40 (const unsigned char *p, const int n)
{
	cob_s64_t	val = 0;
	unsigned char	*x;

	x = (unsigned char *)&val;
	COB_LD_ST5 (x, p);
	val = COB_BSWAP_64 (val);
	val >>= 24;	/* Shift with sign */
	return (val < n) ? -1 : (val > n);
}
#endif

#ifdef	COB_CMPSWP_U48
static COB_INLINE COB_A_INLINE int
cob_cmpswp_u48 (const unsigned char *p, const int n)
{
	cob_u64_t	val = 0;
	unsigned char	*x;

	if (unlikely(n < 0)) {
		return 1;
	}
	x = ((unsigned char *)&val) + 2;
	COB_LD_ST6 (x, p);
	val = COB_BSWAP_64 (val);
	return (val < n) ? -1 : (val > n);
}
#endif

#ifdef	COB_CMPSWP_S48
static COB_INLINE COB_A_INLINE int
cob_cmpswp_s48 (const unsigned char *p, const int n)
{
	cob_s64_t	val = 0;
	unsigned char	*x;

	x = (unsigned char *)&val;
	COB_LD_ST6 (x, p);
	val = COB_BSWAP_64 (val);
	val >>= 16;	/* Shift with sign */
	return (val < n) ? -1 : (val > n);
}
#endif

#ifdef	COB_CMPSWP_U56
static COB_INLINE COB_A_INLINE int
cob_cmpswp_u56 (const unsigned char *p, const int n)
{
	cob_u64_t	val = 0;
	unsigned char	*x;

	if (unlikely(n < 0)) {
		return 1;
	}
	x = ((unsigned char *)&val) + 1;
	COB_LD_ST7 (x, p);
	val = COB_BSWAP_64 (val);
	return (val < n) ? -1 : (val > n);
}
#endif

#ifdef	COB_CMPSWP_S56
static COB_INLINE COB_A_INLINE int
cob_cmpswp_s56 (const unsigned char *p, const int n)
{
	cob_s64_t	val = 0;
	unsigned char	*x;

	x = (unsigned char *)&val;
	COB_LD_ST7 (x, p);
	val = COB_BSWAP_64 (val);
	val >>= 8;	/* Shift with sign */
	return (val < n) ? -1 : (val > n);
}
#endif

#ifdef	COB_CMPSWP_U64
static COB_INLINE COB_A_INLINE int
cob_cmpswp_u64 (const unsigned char *p, const int n)
{
#ifndef COB_ALLOW_UNALIGNED
	unsigned char		*x;
#endif
	cob_u64_t	val;

	if (unlikely(n < 0)) {
		return 1;
	}
#ifdef	COB_ALLOW_UNALIGNED
	val = COB_BSWAP_64 (*(const cob_u64_t __unaligned *)p);
#else
	x = (unsigned char *)&val;
	COB_LD_ST8 (x, p);
	val = COB_BSWAP_64 (val);
#endif
	return (val < n) ? -1 : (val > n);
}
#endif

#ifdef	COB_CMPSWP_S64
static COB_INLINE COB_A_INLINE int
cob_cmpswp_s64 (const unsigned char *p, const int n)
{
	cob_s64_t	val;

#ifdef	COB_ALLOW_UNALIGNED
	val = COB_BSWAP_64 (*(const cob_s64_t __unaligned *)p);
#else
	unsigned char	*x;
	x = (unsigned char *)&val;
	COB_LD_ST8 (x, p);
	val = COB_BSWAP_64 (val);
#endif
	return (val < n) ? -1 : (val > n);
}
#endif

/* Binary swapped add */
#ifdef	COB_ADDSWP_U16
static COB_INLINE COB_A_INLINE void
cob_addswp_u16 (unsigned char *p, const int val)
{
	unsigned short	n;

#ifdef	COB_ALLOW_UNALIGNED
	n = COB_BSWAP_16 (*(unsigned short __unaligned *)p);
	n += val;
	*(unsigned short __unaligned *)p = COB_BSWAP_16(n);
#else
	unsigned char	*x;

	x = (unsigned char *)&n;
	x[0] = p[1];
	x[1] = p[0];
	n += val;
	p[0] = x[1];
	p[1] = x[0];
#endif
}
#endif

#ifdef	COB_ADDSWP_S16
static COB_INLINE COB_A_INLINE void
cob_addswp_s16 (unsigned char *p, const int val)
{
	short		n;

#ifdef	COB_ALLOW_UNALIGNED
	n = COB_BSWAP_16 (*(short __unaligned *)p);
	n += val;
	*(short __unaligned *)p = COB_BSWAP_16(n);
#else
	unsigned char	*x;

	x = (unsigned char *)&n;
	x[0] = p[1];
	x[1] = p[0];
	n += val;
	p[0] = x[1];
	p[1] = x[0];
#endif
}
#endif

#ifdef	COB_ADDSWP_U24
static COB_INLINE COB_A_INLINE void
cob_addswp_u24 (unsigned char *p, const int val)
{
	unsigned char	*x;
	unsigned int	n = 0;

	x = (unsigned char *)&n;
	x[0] = p[2];
	x[1] = p[1];
	x[2] = p[0];
	n += val;
	p[0] = x[2];
	p[1] = x[1];
	p[2] = x[0];
}
#endif

#ifdef	COB_ADDSWP_S24
static COB_INLINE COB_A_INLINE void
cob_addswp_s24 (unsigned char *p, const int val)
{
	unsigned char	*x;
	int		n = 0;

	x = ((unsigned char *)&n) + 1;
	x[0] = p[2];
	x[1] = p[1];
	x[2] = p[0];
	n >>= 8;	/* Shift with sign */
	n += val;
	x = (unsigned char *)&n;
	p[0] = x[2];
	p[1] = x[1];
	p[2] = x[0];
}
#endif

#ifdef	COB_ADDSWP_U32
static COB_INLINE COB_A_INLINE void
cob_addswp_u32 (unsigned char *p, const int val)
{
	unsigned int	n;

#ifdef	COB_ALLOW_UNALIGNED
	n = COB_BSWAP_32 (*(unsigned int __unaligned *)p);
	n += val;
	*(unsigned int __unaligned *)p = COB_BSWAP_32(n);
#else
	unsigned char	*x;

	x = (unsigned char *)&n;
	x[0] = p[3];
	x[1] = p[2];
	x[2] = p[1];
	x[3] = p[0];
	n += val;
	p[0] = x[3];
	p[1] = x[2];
	p[2] = x[1];
	p[3] = x[0];
#endif
}
#endif

#ifdef	COB_ADDSWP_S32
static COB_INLINE COB_A_INLINE void
cob_addswp_s32 (unsigned char *p, const int val)
{
	int		n;

#ifdef	COB_ALLOW_UNALIGNED
	n = COB_BSWAP_32 (*(int __unaligned *)p);
	n += val;
	*(int __unaligned *)p = COB_BSWAP_32(n);
#else
	unsigned char	*x;

	x = (unsigned char *)&n;
	x[0] = p[3];
	x[1] = p[2];
	x[2] = p[1];
	x[3] = p[0];
	n += val;
	p[0] = x[3];
	p[1] = x[2];
	p[2] = x[1];
	p[3] = x[0];
#endif
}
#endif

#ifdef	COB_ADDSWP_U40
static COB_INLINE COB_A_INLINE void
cob_addswp_u40 (unsigned char *p, const int val)
{
	cob_u64_t		n = 0;
	unsigned char		*x;

	x = (unsigned char *)&n;
	x[0] = p[4];
	x[1] = p[3];
	x[2] = p[2];
	x[3] = p[1];
	x[4] = p[0];
	n += val;
	p[0] = x[4];
	p[1] = x[3];
	p[2] = x[2];
	p[3] = x[1];
	p[4] = x[0];
}
#endif

#ifdef	COB_ADDSWP_S40
static COB_INLINE COB_A_INLINE void
cob_addswp_s40 (unsigned char *p, const int val)
{
	cob_s64_t		n = 0;
	unsigned char		*x;

	x = ((unsigned char *)&n) + 3;
	x[0] = p[4];
	x[1] = p[3];
	x[2] = p[2];
	x[3] = p[1];
	x[4] = p[0];
	n >>= 24;	/* Shift with sign */
	n += val;
	x = (unsigned char *)&n;
	p[0] = x[4];
	p[1] = x[3];
	p[2] = x[2];
	p[3] = x[1];
	p[4] = x[0];
}
#endif

#ifdef	COB_ADDSWP_U48
static COB_INLINE COB_A_INLINE void
cob_addswp_u48 (unsigned char *p, const int val)
{
	cob_u64_t		n = 0;
	unsigned char		*x;

	x = (unsigned char *)&n;
	x[0] = p[5];
	x[1] = p[4];
	x[2] = p[3];
	x[3] = p[2];
	x[4] = p[1];
	x[5] = p[0];
	n += val;
	p[0] = x[5];
	p[1] = x[4];
	p[2] = x[3];
	p[3] = x[2];
	p[4] = x[1];
	p[5] = x[0];
}
#endif

#ifdef	COB_ADDSWP_S48
static COB_INLINE COB_A_INLINE void
cob_addswp_s48 (unsigned char *p, const int val)
{
	cob_s64_t		n = 0;
	unsigned char		*x;

	x = ((unsigned char *)&n) + 2;
	x[0] = p[5];
	x[1] = p[4];
	x[2] = p[3];
	x[3] = p[2];
	x[4] = p[1];
	x[5] = p[0];
	n >>= 16;	/* Shift with sign */
	n += val;
	x = (unsigned char *)&n;
	p[0] = x[5];
	p[1] = x[4];
	p[2] = x[3];
	p[3] = x[2];
	p[4] = x[1];
	p[5] = x[0];
}
#endif

#ifdef	COB_ADDSWP_U56
static COB_INLINE COB_A_INLINE void
cob_addswp_u56 (unsigned char *p, const int val)
{
	cob_u64_t		n = 0;
	unsigned char		*x;

	x = (unsigned char *)&n;
	x[0] = p[6];
	x[1] = p[5];
	x[2] = p[4];
	x[3] = p[3];
	x[4] = p[2];
	x[5] = p[1];
	x[6] = p[0];
	n += val;
	p[0] = x[6];
	p[1] = x[5];
	p[2] = x[4];
	p[3] = x[3];
	p[4] = x[2];
	p[5] = x[1];
	p[6] = x[0];
}
#endif

#ifdef	COB_ADDSWP_S56
static COB_INLINE COB_A_INLINE void
cob_addswp_s56 (unsigned char *p, const int val)
{
	cob_s64_t		n = 0;
	unsigned char		*x;

	x = ((unsigned char *)&n) + 1;
	x[0] = p[6];
	x[1] = p[5];
	x[2] = p[4];
	x[3] = p[3];
	x[4] = p[2];
	x[5] = p[1];
	x[6] = p[0];
	n >>= 8;	/* Shift with sign */
	n += val;
	x = (unsigned char *)&n;
	p[0] = x[6];
	p[1] = x[5];
	p[2] = x[4];
	p[3] = x[3];
	p[4] = x[2];
	p[5] = x[1];
	p[6] = x[0];
}
#endif

#ifdef	COB_ADDSWP_U64
static COB_INLINE COB_A_INLINE void
cob_addswp_u64 (unsigned char *p, const int val)
{
	cob_u64_t		n;

#ifdef	COB_ALLOW_UNALIGNED
	n = COB_BSWAP_64 (*(cob_u64_t __unaligned *)p);
	n += val;
	*(cob_u64_t __unaligned *)p = COB_BSWAP_64(n);
#else
	unsigned char		*x;

	x = (unsigned char *)&n;
	x[0] = p[7];
	x[1] = p[6];
	x[2] = p[5];
	x[3] = p[4];
	x[4] = p[3];
	x[5] = p[2];
	x[6] = p[1];
	x[7] = p[0];
	n += val;
	p[0] = x[7];
	p[1] = x[6];
	p[2] = x[5];
	p[3] = x[4];
	p[4] = x[3];
	p[5] = x[2];
	p[6] = x[1];
	p[7] = x[0];
#endif
}
#endif

#ifdef	COB_ADDSWP_S64
static COB_INLINE COB_A_INLINE void
cob_addswp_s64 (unsigned char *p, const int val)
{
	cob_s64_t		n;

#ifdef	COB_ALLOW_UNALIGNED
	n = COB_BSWAP_64 (*(cob_s64_t __unaligned *)p);
	n += val;
	*(cob_s64_t __unaligned *)p = COB_BSWAP_64(n);
#else
	unsigned char		*x;

	x = (unsigned char *)&n;
	x[0] = p[7];
	x[1] = p[6];
	x[2] = p[5];
	x[3] = p[4];
	x[4] = p[3];
	x[5] = p[2];
	x[6] = p[1];
	x[7] = p[0];
	n += val;
	p[0] = x[7];
	p[1] = x[6];
	p[2] = x[5];
	p[3] = x[4];
	p[4] = x[3];
	p[5] = x[2];
	p[6] = x[1];
	p[7] = x[0];
#endif
}
#endif

/* Binary swapped subtract */
#ifdef	COB_SUBSWP_U16
static COB_INLINE COB_A_INLINE void
cob_subswp_u16 (unsigned char *p, const int val)
{
	unsigned short	n;

#ifdef	COB_ALLOW_UNALIGNED
	n = COB_BSWAP_16 (*(unsigned short __unaligned *)p);
	n -= val;
	*(unsigned short __unaligned *)p = COB_BSWAP_16(n);
#else
	unsigned char	*x;

	x = (unsigned char *)&n;
	x[0] = p[1];
	x[1] = p[0];
	n -= val;
	p[0] = x[1];
	p[1] = x[0];
#endif
}
#endif

#ifdef	COB_SUBSWP_S16
static COB_INLINE COB_A_INLINE void
cob_subswp_s16 (unsigned char *p, const int val)
{
	short		n;

#ifdef	COB_ALLOW_UNALIGNED
	n = COB_BSWAP_16 (*(short __unaligned *)p);
	n -= val;
	*(short __unaligned *)p = COB_BSWAP_16(n);
#else
	unsigned char	*x;

	x = (unsigned char *)&n;
	x[0] = p[1];
	x[1] = p[0];
	n -= val;
	p[0] = x[1];
	p[1] = x[0];
#endif
}
#endif

#ifdef	COB_SUBSWP_U24
static COB_INLINE COB_A_INLINE void
cob_subswp_u24 (unsigned char *p, const int val)
{
	unsigned char	*x;
	unsigned int	n = 0;

	x = (unsigned char *)&n;
	x[0] = p[2];
	x[1] = p[1];
	x[2] = p[0];
	n -= val;
	p[0] = x[2];
	p[1] = x[1];
	p[2] = x[0];
}
#endif

#ifdef	COB_SUBSWP_S24
static COB_INLINE COB_A_INLINE void
cob_subswp_s24 (unsigned char *p, const int val)
{
	unsigned char	*x;
	int		n = 0;

	x = ((unsigned char *)&n) + 1;
	x[0] = p[2];
	x[1] = p[1];
	x[2] = p[0];
	n >>= 8;	/* Shift with sign */
	n -= val;
	x = (unsigned char *)&n;
	p[0] = x[2];
	p[1] = x[1];
	p[2] = x[0];
}
#endif

#ifdef	COB_SUBSWP_U32
static COB_INLINE COB_A_INLINE void
cob_subswp_u32 (unsigned char *p, const int val)
{
	unsigned int	n;

#ifdef	COB_ALLOW_UNALIGNED
	n = COB_BSWAP_32 (*(unsigned int __unaligned *)p);
	n -= val;
	*(unsigned int __unaligned *)p = COB_BSWAP_32(n);
#else
	unsigned char	*x;

	x = (unsigned char *)&n;
	x[0] = p[3];
	x[1] = p[2];
	x[2] = p[1];
	x[3] = p[0];
	n -= val;
	p[0] = x[3];
	p[1] = x[2];
	p[2] = x[1];
	p[3] = x[0];
#endif
}
#endif

#ifdef	COB_SUBSWP_S32
static COB_INLINE COB_A_INLINE void
cob_subswp_s32 (unsigned char *p, const int val)
{
	int		n;

#ifdef	COB_ALLOW_UNALIGNED
	n = COB_BSWAP_32 (*(int __unaligned *)p);
	n -= val;
	*(int __unaligned *)p = COB_BSWAP_32(n);
#else
	unsigned char	*x;

	x = (unsigned char *)&n;
	x[0] = p[3];
	x[1] = p[2];
	x[2] = p[1];
	x[3] = p[0];
	n -= val;
	p[0] = x[3];
	p[1] = x[2];
	p[2] = x[1];
	p[3] = x[0];
#endif
}
#endif

#ifdef	COB_SUBSWP_U40
static COB_INLINE COB_A_INLINE void
cob_subswp_u40 (unsigned char *p, const int val)
{
	cob_u64_t		n = 0;
	unsigned char		*x;

	x = (unsigned char *)&n;
	x[0] = p[4];
	x[1] = p[3];
	x[2] = p[2];
	x[3] = p[1];
	x[4] = p[0];
	n -= val;
	p[0] = x[4];
	p[1] = x[3];
	p[2] = x[2];
	p[3] = x[1];
	p[4] = x[0];
}
#endif

#ifdef	COB_SUBSWP_S40
static COB_INLINE COB_A_INLINE void
cob_subswp_s40 (unsigned char *p, const int val)
{
	cob_s64_t		n = 0;
	unsigned char		*x;

	x = ((unsigned char *)&n) + 3;
	x[0] = p[4];
	x[1] = p[3];
	x[2] = p[2];
	x[3] = p[1];
	x[4] = p[0];
	n >>= 24;	/* Shift with sign */
	n -= val;
	x = (unsigned char *)&n;
	p[0] = x[4];
	p[1] = x[3];
	p[2] = x[2];
	p[3] = x[1];
	p[4] = x[0];
}
#endif

#ifdef	COB_SUBSWP_U48
static COB_INLINE COB_A_INLINE void
cob_subswp_u48 (unsigned char *p, const int val)
{
	cob_u64_t		n = 0;
	unsigned char		*x;

	x = (unsigned char *)&n;
	x[0] = p[5];
	x[1] = p[4];
	x[2] = p[3];
	x[3] = p[2];
	x[4] = p[1];
	x[5] = p[0];
	n -= val;
	p[0] = x[5];
	p[1] = x[4];
	p[2] = x[3];
	p[3] = x[2];
	p[4] = x[1];
	p[5] = x[0];
}
#endif

#ifdef	COB_SUBSWP_S48
static COB_INLINE COB_A_INLINE void
cob_subswp_s48 (unsigned char *p, const int val)
{
	cob_s64_t		n = 0;
	unsigned char		*x;

	x = ((unsigned char *)&n) + 2;
	x[0] = p[5];
	x[1] = p[4];
	x[2] = p[3];
	x[3] = p[2];
	x[4] = p[1];
	x[5] = p[0];
	n >>= 16;	/* Shift with sign */
	n -= val;
	x = (unsigned char *)&n;
	p[0] = x[5];
	p[1] = x[4];
	p[2] = x[3];
	p[3] = x[2];
	p[4] = x[1];
	p[5] = x[0];
}
#endif

#ifdef	COB_SUBSWP_U56
static COB_INLINE COB_A_INLINE void
cob_subswp_u56 (unsigned char *p, const int val)
{
	cob_u64_t		n = 0;
	unsigned char		*x;

	x = (unsigned char *)&n;
	x[0] = p[6];
	x[1] = p[5];
	x[2] = p[4];
	x[3] = p[3];
	x[4] = p[2];
	x[5] = p[1];
	x[6] = p[0];
	n -= val;
	p[0] = x[6];
	p[1] = x[5];
	p[2] = x[4];
	p[3] = x[3];
	p[4] = x[2];
	p[5] = x[1];
	p[6] = x[0];
}
#endif

#ifdef	COB_SUBSWP_S56
static COB_INLINE COB_A_INLINE void
cob_subswp_s56 (unsigned char *p, const int val)
{
	cob_s64_t		n = 0;
	unsigned char		*x;

	x = ((unsigned char *)&n) + 1;
	x[0] = p[6];
	x[1] = p[5];
	x[2] = p[4];
	x[3] = p[3];
	x[4] = p[2];
	x[5] = p[1];
	x[6] = p[0];
	n >>= 8;	/* Shift with sign */
	n -= val;
	x = (unsigned char *)&n;
	p[0] = x[6];
	p[1] = x[5];
	p[2] = x[4];
	p[3] = x[3];
	p[4] = x[2];
	p[5] = x[1];
	p[6] = x[0];
}
#endif

#ifdef	COB_SUBSWP_U64
static COB_INLINE COB_A_INLINE void
cob_subswp_u64 (unsigned char *p, const int val)
{
	cob_u64_t		n;

#ifdef	COB_ALLOW_UNALIGNED
	n = COB_BSWAP_64 (*(cob_u64_t __unaligned *)p);
	n -= val;
	*(cob_u64_t __unaligned *)p = COB_BSWAP_64(n);
#else
	unsigned char		*x;

	x = (unsigned char *)&n;
	x[0] = p[7];
	x[1] = p[6];
	x[2] = p[5];
	x[3] = p[4];
	x[4] = p[3];
	x[5] = p[2];
	x[6] = p[1];
	x[7] = p[0];
	n -= val;
	p[0] = x[7];
	p[1] = x[6];
	p[2] = x[5];
	p[3] = x[4];
	p[4] = x[3];
	p[5] = x[2];
	p[6] = x[1];
	p[7] = x[0];
#endif
}
#endif

#ifdef	COB_SUBSWP_S64
static COB_INLINE COB_A_INLINE void
cob_subswp_s64 (unsigned char *p, const int val)
{
	cob_s64_t		n;

#ifdef	COB_ALLOW_UNALIGNED
	n = COB_BSWAP_64 (*(cob_s64_t __unaligned *)p);
	n -= val;
	*(cob_s64_t __unaligned *)p = COB_BSWAP_64(n);
#else
	unsigned char		*x;

	x = (unsigned char *)&n;
	x[0] = p[7];
	x[1] = p[6];
	x[2] = p[5];
	x[3] = p[4];
	x[4] = p[3];
	x[5] = p[2];
	x[6] = p[1];
	x[7] = p[0];
	n -= val;
	p[0] = x[7];
	p[1] = x[6];
	p[2] = x[5];
	p[3] = x[4];
	p[4] = x[3];
	p[5] = x[2];
	p[6] = x[1];
	p[7] = x[0];
#endif
}
#endif

#ifdef	COB_SETSWP_U16
static COB_INLINE COB_A_INLINE void
cob_setswp_u16 (unsigned char *p, const int val)
{
	unsigned short	n;

	n = val;
#ifdef	COB_ALLOW_UNALIGNED
	*(unsigned short __unaligned *)p = COB_BSWAP_16(n);
#else
	unsigned char	*x;

	x = (unsigned char *)&n;
	p[0] = x[1];
	p[1] = x[0];
#endif
}
#endif

#ifdef	COB_SETSWP_S16
static COB_INLINE COB_A_INLINE void
cob_setswp_s16 (unsigned char *p, const int val)
{
	short		n;

	n = val;
#ifdef	COB_ALLOW_UNALIGNED
	*(short __unaligned *)p = COB_BSWAP_16(n);
#else
	unsigned char	*x;

	x = (unsigned char *)&n;
	p[0] = x[1];
	p[1] = x[0];
#endif
}
#endif

#ifdef	COB_SETSWP_U24
static COB_INLINE COB_A_INLINE void
cob_setswp_u24 (unsigned char *p, const int val)
{
	unsigned char	*x;
	unsigned int	n;

	n = val;
	x = (unsigned char *)&n;
	p[0] = x[2];
	p[1] = x[1];
	p[2] = x[0];
}
#endif

#ifdef	COB_SETSWP_S24
static COB_INLINE COB_A_INLINE void
cob_setswp_s24 (unsigned char *p, const int val)
{
	unsigned char	*x;
	int		n;

	n = val;
	x = (unsigned char *)&n;
	p[0] = x[2];
	p[1] = x[1];
	p[2] = x[0];
}
#endif

#ifdef	COB_SETSWP_U32
static COB_INLINE COB_A_INLINE void
cob_setswp_u32 (unsigned char *p, const int val)
{
	unsigned int	n;

	n = val;
#ifdef	COB_ALLOW_UNALIGNED
	*(unsigned int __unaligned *)p = COB_BSWAP_32(n);
#else
	unsigned char		*x;

	x = (unsigned char *)&n;
	p[0] = x[3];
	p[1] = x[2];
	p[2] = x[1];
	p[3] = x[0];
#endif
}
#endif

#ifdef	COB_SETSWP_S32
static COB_INLINE COB_A_INLINE void
cob_setswp_s32 (unsigned char *p, const int val)
{
	int		n;

	n = val;
#ifdef	COB_ALLOW_UNALIGNED
	*(int __unaligned *)p = COB_BSWAP_32(n);
#else
	unsigned char		*x;

	x = (unsigned char *)&n;
	p[0] = x[3];
	p[1] = x[2];
	p[2] = x[1];
	p[3] = x[0];
#endif
}
#endif

#ifdef	COB_SETSWP_U40
static COB_INLINE COB_A_INLINE void
cob_setswp_u40 (unsigned char *p, const int val)
{
	cob_u64_t		n;
	unsigned char		*x;

	n = val;
	x = (unsigned char *)&n;
	p[0] = x[4];
	p[1] = x[3];
	p[2] = x[2];
	p[3] = x[1];
	p[4] = x[0];
}
#endif

#ifdef	COB_SETSWP_S40
static COB_INLINE COB_A_INLINE void
cob_setswp_s40 (unsigned char *p, const int val)
{
	cob_s64_t		n;
	unsigned char		*x;

	n = val;
	x = (unsigned char *)&n;
	p[0] = x[4];
	p[1] = x[3];
	p[2] = x[2];
	p[3] = x[1];
	p[4] = x[0];
}
#endif

#ifdef	COB_SETSWP_U48
static COB_INLINE COB_A_INLINE void
cob_setswp_u48 (unsigned char *p, const int val)
{
	cob_u64_t		n;
	unsigned char		*x;

	n = val;
	x = (unsigned char *)&n;
	p[0] = x[5];
	p[1] = x[4];
	p[2] = x[3];
	p[3] = x[2];
	p[4] = x[1];
	p[5] = x[0];
}
#endif

#ifdef	COB_SETSWP_S48
static COB_INLINE COB_A_INLINE void
cob_setswp_s48 (unsigned char *p, const int val)
{
	cob_s64_t		n;
	unsigned char		*x;

	n = val;
	x = (unsigned char *)&n;
	p[0] = x[5];
	p[1] = x[4];
	p[2] = x[3];
	p[3] = x[2];
	p[4] = x[1];
	p[5] = x[0];
}
#endif

#ifdef	COB_SETSWP_U56
static COB_INLINE COB_A_INLINE void
cob_setswp_u56 (unsigned char *p, const int val)
{
	cob_u64_t		n;
	unsigned char		*x;

	n = val;
	x = (unsigned char *)&n;
	p[0] = x[6];
	p[1] = x[5];
	p[2] = x[4];
	p[3] = x[3];
	p[4] = x[2];
	p[5] = x[1];
	p[6] = x[0];
}
#endif

#ifdef	COB_SETSWP_S56
static COB_INLINE COB_A_INLINE void
cob_setswp_s56 (unsigned char *p, const int val)
{
	cob_s64_t		n;
	unsigned char		*x;

	n = val;
	x = (unsigned char *)&n;
	p[0] = x[6];
	p[1] = x[5];
	p[2] = x[4];
	p[3] = x[3];
	p[4] = x[2];
	p[5] = x[1];
	p[6] = x[0];
}
#endif

#ifdef	COB_SETSWP_U64
static COB_INLINE COB_A_INLINE void
cob_setswp_u64 (unsigned char *p, const int val)
{
	cob_u64_t		n;

	n = val;
#ifdef	COB_ALLOW_UNALIGNED
	*(cob_u64_t __unaligned *)p = COB_BSWAP_64(n);
#else
	unsigned char		*x;

	x = (unsigned char *)&n;
	p[0] = x[7];
	p[1] = x[6];
	p[2] = x[5];
	p[3] = x[4];
	p[4] = x[3];
	p[5] = x[2];
	p[6] = x[1];
	p[7] = x[0];
#endif
}
#endif

#ifdef	COB_SETSWP_S64
static COB_INLINE COB_A_INLINE void
cob_setswp_s64 (unsigned char *p, const int val)
{
	cob_s64_t		n;

	n = val;
#ifdef	COB_ALLOW_UNALIGNED
	*(cob_s64_t __unaligned *)p = COB_BSWAP_64(n);
#else
	unsigned char		*x;

	x = (unsigned char *)&n;
	p[0] = x[7];
	p[1] = x[6];
	p[2] = x[5];
	p[3] = x[4];
	p[4] = x[3];
	p[5] = x[2];
	p[6] = x[1];
	p[7] = x[0];
#endif
}
#endif

#undef	COB_LD_ST1
#undef	COB_LD_ST2
#undef	COB_LD_ST3
#undef	COB_LD_ST4
#undef	COB_LD_ST5
#undef	COB_LD_ST6
#undef	COB_LD_ST7
#undef	COB_LD_ST8

#endif	/* COB_CODEGEN_H */
